"""Sets up the communication for the gas card device.

Author: Grayson Bellamy
Date: 2024-01-05
"""

from typing import Optional
from collections.abc import ByteString

from abc import ABC, abstractmethod

import trio
from trio_serial import Parity, SerialStream, StopBits


class CommDevice(ABC):
    """Sets up the communication for the an Alicat device."""

    def __init__(self, timeout: int) -> None:
        """Initializes the serial communication.

        Args:
            timeout (int): The timeout of the Alicat device.
        """
        self.timeout = timeout

    @abstractmethod
    async def _read(self, len: int) -> Optional[str]:
        """Reads the serial communication.

        Args:
            len (int): The length of the serial communication to read. One character if not specified.

        Returns:
            str: The serial communication.
        """
        pass

    @abstractmethod
    async def _write(self, command: str) -> None:
        """Writes the serial communication.

        Args:
            command (str): The serial communication.
        """
        pass

    @abstractmethod
    async def close(self):
        """Closes the serial communication."""
        pass

    @abstractmethod
    async def _readline(self) -> Optional[str]:
        """Reads the serial communication until end-of-line character reached.

        Returns:
            str: The serial communication.
        """
        pass

    @abstractmethod
    async def _write_readline(self, command: str) -> Optional[str]:
        """Writes the serial communication and reads the response until end-of-line character reached.

        Args:
            command (str): The serial communication.

        Returns:
            str: The serial communication.
        """
        pass


class SerialDevice(CommDevice):
    """Sets up the communication for the a gas card device using serial protocol."""

    def __init__(  # Need to verify Flow Control is set to 'none'
        self,
        port: str,
        baudrate: int = 57600,
        timeout: int = 150,  # Not present in manual
        databits: int = 8,
        parity: Parity = Parity.NONE,
        stopbits: StopBits = StopBits.ONE,
        xonxoff: bool = False,  # Not present in manual
        rtscts: bool = False,  # Not present in manual
        exclusive: bool = False,  # Not present in manual
    ):
        """Initializes the serial communication.

        Args:
            port (str): The port to which the Alicat device is connected.
            baudrate (int): The baudrate of the Alicat device.
            timeout (int): The timeout of the Alicat device in ms.
            databits (int): The number of data bits.
            parity (Parity): The parity of the Alicat device.
            stopbits (StopBits): The of stop bits. Usually 1 or 2.
            xonxoff (bool): Whether the port uses xonxoff.
            rtscts (bool): Whether the port uses rtscts.
            exclusive (bool): Whether the port is exclusive.
        """
        super().__init__(timeout)

        self.timeout = timeout
        self.eol = b"\r"
        self.serial_setup = {
            "port": port,
            "exclusive": exclusive,
            "baudrate": baudrate,
            "bytesize": databits,
            "parity": parity,
            "stopbits": stopbits,
            "xonxoff": xonxoff,
            "rtscts": rtscts,
        }
        self.isOpen = False
        self.ser_devc = SerialStream(**self.serial_setup)

    async def _read(self, len: int = 1) -> ByteString:
        """Reads the serial communication.

        Args:
            len (int): The length of the serial communication to read. One character if not specified.

        Returns:
            ByteString: The serial communication.
        """
        if not self.isOpen:
            async with self.ser_devc:
                with trio.move_on_after(self.timeout / 1000):
                    return await self.ser_devc.receive_some(len)
        else:
            with trio.move_on_after(self.timeout / 1000):
                return await self.ser_devc.receive_some(len)
        return None

    async def _write(self, command: str) -> None:
        """Writes the serial communication.

        Args:
            command (str): The serial communication.
        """
        if not self.isOpen:
            async with self.ser_devc:
                with trio.move_on_after(self.timeout / 1000):
                    await self.ser_devc.send_all(command.encode("ascii") + self.eol)
        else:
            with trio.move_on_after(self.timeout / 1000):
                await self.ser_devc.send_all(command.encode("ascii") + self.eol)
        return None

    async def _readline(self) -> str:
        """Reads the serial communication until end-of-line character reached.

        Returns:
            str: The serial communication.
        """
        async with self.ser_devc:
            self.isOpen = True
            line = bytearray()
            c = None
            while (
                c != self.eol
            ):  # Keep reading until end-of-line character reached, then we know new line is started
                c = await self._read(1)
            while True:
                c = None
                with trio.move_on_after(self.timeout / 1000):
                    c = await self._read(1)
                    line += c
                    if c == self.eol:
                        break
                if c is None:
                    break
        self.isOpen = False
        return line.decode("ascii")

    async def _write_readall(self, command: str) -> list:
        """Write command and read until timeout reached.

        Args:
            command (str): The serial communication.

        Returns:
            list: List of lines read from the device.
        """
        async with self.ser_devc:
            self.isOpen = True
            await self._write(command)
            line = bytearray()
            arr_line = []
            await self._flush()
            while True:
                c = None
                with trio.move_on_after(self.timeout / 1000):
                    c = await self._read(1)
                    if c == self.eol:
                        arr_line.append(line.decode("ascii"))
                        line = bytearray()
                    else:
                        line += c
                if c is None:
                    break
        self.isOpen = False
        return arr_line

    async def _write_readline(self, command: str) -> str:
        """Writes the serial communication and reads the response until end-of-line character reached.

        Parameters:
            command (str): The serial communication.

        Returns:
            str: The serial communication.
        """
        async with self.ser_devc:
            self.isOpen = True
            await self._write(command)
            line = bytearray()
            c = None
            while (
                c != self.eol and c is not None
            ):  # Keep reading until end-of-line character reached, then we know new line is started
                c = await self._read(1)
            while True:
                c = None
                with trio.move_on_after(self.timeout / 1000):
                    c = await self._read(1)
                    if c == self.eol:
                        break
                    line += c
                if c is None:
                    break
        self.isOpen = False
        return line.decode("ascii")

    async def _flush(self) -> None:
        """Flushes the serial communication."""
        await self.ser_devc.discard_input()

    async def close(self) -> None:
        """Closes the serial communication."""
        self.isOpen = False
        await self.ser_devc.aclose()

    async def open(self) -> None:
        """Opens the serial communication."""
        self.isOpen = True
        await self.ser_devc.aopen()



"""DAQ Class for managing gascard devices. Accessible to external API and internal logging module.

Author: Grayson Bellamy
Date: 2024-01-07
"""

import device
from trio import run


class DAQ:
    """Class for managing gascard devices. Accessible to external API and internal logging module. Wraps and allows communication with inidividual or all devices through wrapper class."""

    def __init__(self) -> None:
        """Initializes the DAQ.

        TODO: Pass dictionary of names and addresses to initialize devices. Same async issue.

        """
        global dev_list
        dev_list = {}

        """
        for name in devs:
            dev = device.new_device(devs[name])
            dev_list.update({name: dev})
        """
        return

    @classmethod
    async def init(cls, devs: dict) -> "DAQ":
        """Initializes the DAQ.

        Args:
            devs (dict): The dictionary of devices to add. Name:Port

        Returns:
            DAQ: The DAQ object.
        """
        daq = cls()
        await daq.add_device(devs)
        return daq

    async def add_device(self, devs: dict) -> None:
        """Creates and initializes the devices.

        Args:
            devs (dict): The dictionary of devices to add. Name:Port
        """
        if isinstance(devs, str):
            devs = devs.split()
            # This works if the string is the format "Name Port"
            devs = {devs[0]: devs[1]}
        for name in devs:
            dev = await device.Gascard.new_device(devs[name])
            dev_list.update({name: dev})
        return

    async def remove_device(self, name: list) -> None:
        """Creates and initializes the devices.

        Args:
            name (list): The list of devices to remove.
        """
        for n in name:
            await dev_list[n]._device.close()
            del dev_list[n]
        return

    async def dev_list(self) -> dict:
        """Displays the list of devices.

        Returns:
            dict: The list of devices and their objects.
        """
        return dev_list

    async def get(self, val: list = "", id: list = "") -> dict:
        """Gets the data from the device.

        If id not specified, returns data from all devices.

        Args:
           val (list): The values to get from the device.
           id (list): The IDs of the devices to read from. If not specified, returns data from all devices.

        Returns:
            dict: The dictionary of devices with the data for each value.
        """
        ret_dict = {}
        if isinstance(val, str):
            val = [val]
        if not id:
            for dev in dev_list:
                ret_dict.update({dev: await dev_list[dev].get(val)})
        if isinstance(id, str):
            id = id.split()
        for i in id:
            ret_dict.update({i: await dev_list[i].get(val)})
        return ret_dict

    async def set(self, command: dict, id: str = "") -> None:
        """Sets the data of the device.

        Args:
           command (dict): The commands and their relevant parameters to send to the device.
           id (list): The IDs of the devices to read from. If not specified, returns data from all devices.

        Returns:
            dict: The dictionary of devices with the data for each value.
        """
        ret_dict = {}
        if isinstance(command, str):
            command = command.split()
        if not id:
            for dev in dev_list:
                ret_dict.update({dev: await dev_list[dev].set(command)})
        if isinstance(id, str):
            id = id.split()
        for i in id:
            ret_dict.update({i: await dev_list[i].set(command)})
        return ret_dict

    async def zero(self, id: str = "") -> None:
        """Sets the zero reference of the device.

        Note:
            **Device MUST be flowing zero gas BEFORE calling this function.**

        Args:
            id (list): The IDs of the devices to read from. If not specified, returns data from all devices.
        """
        ret_dict = {}
        if not id:
            for dev in dev_list:
                ret_dict.update(
                    {dev: await dev_list[i].set({"Zero Gas Corr Factor": ""})}
                )
        if isinstance(id, str):
            id = id.split()
        for i in id:
            ret_dict.update({i: await dev_list[i].set({"Zero Gas Corr Factor": ""})})
        return ret_dict

    async def span(self, val: float, id: str = "") -> None:
        """Sets the span reference of the device.

        Note:
            **Device MUST be flowing span gas BEFORE calling this function.**

        Args:
            val (dict): Gas concentration as a fraction of full scale (0.5 to 1.2)
            id (list): The IDs of the devices to read from. If not specified, returns data from all devices.
        """
        ret_dict = {}
        if not id:
            for dev in dev_list:
                ret_dict.update(
                    {dev: await dev_list[i].set({"Span Gas Corr Factor": val})}
                )
        if isinstance(id, str):
            id = id.split()
        for i in id:
            ret_dict.update({i: await dev_list[i].set({"Span Gas Corr Factor": val})})
        return ret_dict

    async def time_const(self, val: int, id: str = "") -> None:
        """Sets the time constant of the RC filter of the device.

        Args:
            val (int): Time constant in seconds (0 to 120)
            id (list): The IDs of the devices to read from. If not specified, returns data from all devices.
        """
        ret_dict = {}
        if not id:
            for dev in dev_list:
                ret_dict.update({dev: await dev_list[i].set({"Time Constant": val})})
        if isinstance(id, str):
            id = id.split()
        for i in id:
            ret_dict.update({i: await dev_list[i].set({"Time Constant": val})})
        return ret_dict


class DAQLogging:
    """Class for logging the data from gascard devices. Creates and saves file to disk with given acquisition rate. Only used for standalone logging. Use external API for use as plugin."""

    def __init__(self, config: dict) -> None:
        """Initializes the Logging module. Creates and saves file to disk with given acquisition rate.

        Parameters
        ----------
        config : dict
            The configuration dictionary. {Name : port}
        """
        pass



"""Utilities for manipulating data from gascard devices.

Author: Grayson Bellamy
Date: 2024-05-03
"""

import glob
from comm import SerialDevice
from device import Gascard
from trio import run
import trio
import re
from typing import Any


def gas_correction():
    """Calculates the gas correction factor for the gascard device.

    Returns:
    -------
    float
        The gas correction factor.
    """
    pass


async def find_devices():
    """Finds all connected gascard devices.

    Find all available serial ports using the `ls` command
    Iterate through all possible baud rates

    If there is an gascard device on that port (copy code form new_device that checks )
        get what device is
    add to dictionary with port name and type of device on port (if no device, don't add to dictionary)
    return dictionary


    Returns:
    -------
    list
        A list of all connected gascard devices.
    """
    # Get the list of available serial ports
    result = glob.glob("/dev/ttyUSB*")

    # Iterate through the output and check for gascard devices
    devices = {}
    for port in result:
        # Check if the port is an gascard device
        dev = await is_gascard_device(port)
        if dev:
            devices.update({port: dev[1]})
    return devices


async def is_gascard_device(port, id: str = "A", **kwargs: Any):
    """Check if the given port is an gascard device.

    Parameters:
    ----------
    port : str
        The name of the serial port.

    Returns:
    -------
    bool
        True if the port is an gascard device, False otherwise.
    """
    if port.startswith("/dev/"):
        device = SerialDevice(port, **kwargs)
    dev_info_raw = await device._write_readline("X")
    if not dev_info_raw:
        return False
    X_labels = [
        "Mode",
        "Firmware Version",
        "Serial Number",
        "Config Register",
        "Frequency",
        "Time Constant",
        "Switches State",
    ]
    dev_info_raw = dev_info_raw.replace("\x00", "")
    dev_info = dict(zip(X_labels, dev_info_raw.split()))
    if dev_info.get("Serial Number", ""):
        return (True, "Gascard")
    return False


def get_device_type(port):
    """Get the device type for the given port.

    Parameters:
    ----------
    port : str
        The name of the serial port.

    Returns:
    -------
    dict
        A dictionary containing the port name and the type of device on the port.
    """
    # Implement the logic to get the device information
    # You can use any method that suits your needs
    pass
